#include <avr/io.h>
#include <util/delay.h>
#include <stdio.h>
#include <stdlib.h>
#include "lcd.h"

#include <avr/pgmspace.h>
#include <avr/interrupt.h>

#define WALL_1	6
#define WALL_2	60

double score = 0;
/* Game states */
typedef enum STATE {LOADING, GAME, EXIT} state_t;
static state_t state = LOADING;

//matricea zonei de joc
unsigned char matrix[6][84];

typedef struct coordonate {
	int x; 
	int y;
} cords;

// structura ce descrie un obiect
typedef struct obiect {
	cords coord1, coord2, coord3, coord4; 
	int mode;
} obiect; 

obiect current_object;

void update_lcd() {	
	int i, j;
	for (i = 1; i < 6; i++) {
		int start = (i - 1) * 84 + WALL_1 + 1;
		lcd_base_addr(start);
		for (j = WALL_1 + 1; j < WALL_2; j++) 
			lcd_send(matrix[i][j], LCD_DATA);
	}
}

void set_pixel(int x, int y, int value)
{
	int row = x / 8;
	int bit = x % 8;
	if (value == 1)
		matrix[row][y] |= (1<<bit);
	else
		matrix[row][y] &= ~(1<<bit);
}

int check_pixel(int x, int y) {
	int row = x / 8;
	int bit = x % 8;
	return matrix[row][y] & (1<<bit);
}

// unitatea de baza a joclui 
void draw_basic_square(int x, int y)
{
	for (int i=0; i < 2; i++) 
		for (int j=0; j < 2; j++) 
			set_pixel(x+i, y+j, 1);
	
}

void delete_basic_square(int x, int y)
{
	for (int i=0; i < 2; i++) 
		for (int j=0; j < 2; j++) 
			set_pixel(x+i, y+j, 0);
	
}

void delete_square(int x, int y) {
	delete_basic_square(x, y); 
	delete_basic_square(x+2, y);
	delete_basic_square(x, y+2);
	delete_basic_square(x+2, y+2);
}


void draw_square(int x, int y) {
	draw_basic_square(x, y); 
	draw_basic_square(x+2, y);
	draw_basic_square(x, y+2);
	draw_basic_square(x+2, y+2);
	
	current_object.coord2.x = x+2;
	current_object.coord2.y = y;
	
	current_object.coord3.x = x;
	current_object.coord3.y = y+2;
	
	current_object.coord4.x = x+2;
	current_object.coord4.y = y+2;
}

void delete_line(int x, int y) {
	delete_basic_square(x, y); 
	delete_basic_square(x+2, y);
	delete_basic_square(x+4, y);
	delete_basic_square(x+6, y);
}

void draw_line(int x, int y) {
	draw_basic_square(x, y); 
	draw_basic_square(x+2, y);
	draw_basic_square(x+4, y);
	draw_basic_square(x+6, y);
	
	current_object.coord2.x = x+2;
	current_object.coord2.y = y;
	
	current_object.coord3.x = x+4;
	current_object.coord3.y = y;
	
	current_object.coord4.x = x+6;
	current_object.coord4.y = y;
}

void draw_rectangle(int x, int y) {
	draw_basic_square(x, y); 
	draw_basic_square(x+2, y);
	
	draw_basic_square(x, y+2);
	draw_basic_square(x+2, y+2);
	
	draw_basic_square(x, y+4);
	draw_basic_square(x+2, y+4);
	
	draw_basic_square(x, y+6);
	draw_basic_square(x+2, y+6);
	
	current_object.coord2.x = x+2;
	current_object.coord2.y = y;
	
	current_object.coord3.x = x;
	current_object.coord3.y = y+6;
	
	current_object.coord4.x = x+2;
	current_object.coord4.y = y+6;
}

void delete_rectangle(int x, int y) {
	delete_basic_square(x, y); 
	delete_basic_square(x+2, y);
	delete_basic_square(x, y+2);
	delete_basic_square(x+2, y+2);
	delete_basic_square(x, y+4);
	delete_basic_square(x+2, y+4);
	delete_basic_square(x, y+6);
	delete_basic_square(x+2, y+6);
}

/* Prints walls*/
void print_scene()
{

	int wall1_column = 1;
	int wall2_column = 11;
	
	// print first wall
	for (int i = 1; i < 7; i++) {
		lcd_goto_xy(wall1_column, i);
		lcd_chr('I');
	}
	
	// print second wall
	for (int i = 1; i < 7; i++) {
		lcd_goto_xy(wall2_column, i);
		lcd_chr('I');
	}
}

void print_level(int level)
{
	lcd_goto_xy(12, 4);
	lcd_chr('L');
	lcd_goto_xy(13, 4);
	lcd_chr('V');
	lcd_goto_xy(14, 4);
	lcd_chr('L');

	lcd_goto_xy(13, 5);
	if (level == 1)
		lcd_chr('1');
	else if (level == 2)
		lcd_chr('2');
	else if (level == 3)
		lcd_chr('3');
}

// print number of lives
void print_lives(int number_of_lives)
{
	int lifes_column = 14;
	if (number_of_lives == 1) {
		lcd_goto_xy(lifes_column, 1);
		lcd_chr('*');
		lcd_goto_xy(lifes_column, 2);
		lcd_chr('X');
		lcd_goto_xy(lifes_column, 3);
		lcd_chr('X');
	}
	if (number_of_lives == 2) {
		lcd_goto_xy(lifes_column, 1);
		lcd_chr('*');
		lcd_goto_xy(lifes_column, 2);
		lcd_chr('*');
		lcd_goto_xy(lifes_column, 3);
		lcd_chr('X');
	}		
	if (number_of_lives == 3) {
		lcd_goto_xy(lifes_column, 1);
		lcd_chr('*');
		lcd_goto_xy(lifes_column, 2);
		lcd_chr('*');
		lcd_goto_xy(lifes_column, 3);
		lcd_chr('*');
	}
}


/* Display welcoming text */
void loading()
{
	lcd_clear();
	lcd_goto_xy(1, 1);
	lcd_chr('*');
	my_lcd_str("Welcome! Press first button to start!", 2000);
	
	while (1) {
		if ((PINA & (1 << PA0)) == 0) {
			// Setup game
			state = GAME;
			return;
		}	
		_delay_ms(50);
	}
	
}

/* Exit the game or restart*/
void exit()
{
	lcd_clear();
	lcd_goto_xy(1, 1);
	if (score <= 100)
		my_lcd_str("Game Over! Press second button to restart! Score: 100", 2000);
	else if (score <= 200)
		my_lcd_str("Game Over! Press second button to restart! Score: 200", 2000);
	else if (score <= 300)
		my_lcd_str("Game Over! Press second button to restart! Score: 300", 2000);	
	else if (score <= 400)
		my_lcd_str("Game Over! Press second button to restart! Score: 400", 2000);	
	else
		my_lcd_str("Game Over! Press second button to restart! Score: 500", 2000);
		
	while (1) {
		if ((PINA & (1 << PA1)) == 0) {
			// Setup game
			state = LOADING;
			lcd_clear();
			return;
		}	
		_delay_ms(50);
	}
}

bool collide(cords coord)
{
	if (check_pixel(coord.x + 4, coord.y))
		return true;
	if (check_pixel(coord.x + 4, coord.y + 1))
		return true;
	return false;
}

bool collide_on_y(cords coord, int sign)
{
	if (check_pixel(coord.x, coord.y + (1 * sign)))
		return true;
	if (check_pixel(coord.x + 1, coord.y + (1 * sign)))
		return true;
	if (check_pixel(coord.x, coord.y + (2 * sign)))
		return true;
	if (check_pixel(coord.x + 1, coord.y + (2 * sign)))
		return true;	
	if (check_pixel(coord.x, coord.y + (3 * sign)))
		return true;
	if (check_pixel(coord.x + 1, coord.y + (3 * sign)))
		return true;
	if (check_pixel(coord.x, coord.y + (4 * sign)))
		return true;
	if (check_pixel(coord.x + 1, coord.y + (4 * sign)))
		return true;	
	return false;
}


/* Logic for GAME state */
void game()
{
	int number_of_lives = 3;
	int level = 1;
	
	lcd_clear();

	print_scene();
	print_level(level);
	print_lives(number_of_lives);
	
	current_object.coord1.x = 8;
	current_object.coord1.y = WALL_1+1;
	
	int form = rand()%3;
	while(1) {
	
		if (current_object.coord1.x == 8) {
			form = rand()%3;
			if (form == 0){	//square
				score += 10;
				current_object.coord1.y = rand()%(50-1) + 1;
			}
			if (form == 1){	//rectangle
				score += 20;
				current_object.coord1.y = rand()%(46-1) + 1;
			}
			if (form == 2){	//line
				score += 30;
				current_object.coord1.y = rand()%(52-1) + 1;
			}
			if (collide(current_object.coord1)) {
				number_of_lives--;
				if (number_of_lives <= 0) {
					state = EXIT;
					return;
				}
				print_lives(number_of_lives);
				continue;
			}
		}
		
		if (form == 0)
			draw_square(current_object.coord1.x, current_object.coord1.y);
		if (form == 1)
			draw_rectangle(current_object.coord1.x, current_object.coord1.y);
		if (form == 2)
			draw_line(current_object.coord1.x, current_object.coord1.y);
		update_lcd();
		_delay_ms(10000);
		
		if (form == 2) {
			if ((current_object.coord1.x != 44) && 
			!collide(current_object.coord4)) {
				delete_line(current_object.coord1.x, current_object.coord1.y);
				update_lcd();
				_delay_ms(10000);
			}
		}
		else if ((current_object.coord1.x != 44) && 
			!collide(current_object.coord2) &&
			!collide(current_object.coord4)) {
			
			if (form == 0)
				delete_square(current_object.coord1.x, current_object.coord1.y);
			if (form == 1)
				delete_rectangle(current_object.coord1.x, current_object.coord1.y);			
			update_lcd();
			_delay_ms(10000);
		}
		
		_delay_ms(1000);
		if ((PINA & (1 << PA2)) == 0) {
			if (current_object.coord1.y > WALL_1 + 4 && !collide_on_y(current_object.coord1, -1) && !collide_on_y(current_object.coord2, -1))
				current_object.coord1.y -= 4;
		}	

		if ((PINA & (1 << PA3)) == 0) {
			if (current_object.coord1.y < WALL_1 + 46 && !collide_on_y(current_object.coord3, 1) && !collide_on_y(current_object.coord4, 1))
				current_object.coord1.y += 4;
		}
		
		if ((current_object.coord1.x <= 40) &&
			!collide(current_object.coord2) &&
			!collide(current_object.coord4))
			
			current_object.coord1.x += 4;
		else
			current_object.coord1.x = 8;
	}
	lcd_clear();
	state =  EXIT;
}


int main(void)
{
	/* Setup LCD */
	lcd_init();
	lcd_contrast(0x40);
	
	/* Setup buttons */
	DDRA &= ~(1 << PA0);
	DDRA &= ~(1 << PA1);
	DDRA &= ~(1 << PA2);
	DDRA &= ~(1 << PA3);
	
	PORTA |= (1 << PA0);
	PORTA |= (1 << PA1);
	PORTA |= (1 << PA2);
	PORTA |= (1 << PA3);
	
	while (1) {
		switch (state) {
		case LOADING:
			loading();
			break;
		case GAME:
			game();
			break;
		case EXIT:
			exit();
		default:
			break;
		}
	}
}